/* Datei: test_regelung.c
 * Autor: gkemnitz
 * letzte Änderung: 28.09.2016
 *
 * Experimentierprogramm für die 
 * Dimensionierung eines PI-Motorreglers
 *
 * Motoren über zwei H-Brücken PModHB5 und Y-Kabel an JL
 * PModUSBUSART an JH und über USB-Kabel mit PC verbinden.
 * LCD (PmodLCS) an JD oben.
 * JHX und JLX gekreuzt.
 * Tasten-Modul (Schalter-Modul) an JA, Einschalten eines 
 * beliebigen Schalters/Tasters bewirkt ein Notaus (sofortigen Stop).
 */

#include "pwm.h"
#include "rotmess.h"
#include "comir_pc.h"
#include "comir_lcd.h"
#include <avr/interrupt.h>
#include <stdlib.h>

// Begrenzung des Betrags für 32-Bit-Zahlen
void limit(int32_t *val, int32_t max){
  if (*val >  max)
    *val = max;
  else if (*val < -max)
    *val = -max;
}
					
// globale private Daten der Regelung
uint8_t stop_ct=1;         // Zähler Bewegungsstopps
uint16_t step_ct;          // Schrittzähler
uint16_t kp, ki;           // Reglerparameter
int32_t diffR, diffL;      // Schrittdifferenz
int32_t integR, integL;    // Integralanteile
int32_t pwmR, pwmL;        // Stellwerte
int16_t speedR, speedL;    // Sollgeschwindigkeiten 
uint8_t mrmsg[COM_PC_RMSG_LEN]; // Empfangsnachricht
uint8_t msmsg[COM_PC_SMSG_LEN]; // Sendenachricht


// Anzeigepositionen auf dem LC-Display
#define INITSTR "A:xxx xxx T:xx .B:xxx xxx E:...."
#define LCP_DIFF_R   2     // Differenz Motor R
#define LCP_INTEG_R  6     // Integralteil Motor R
#define LCP_STPCT   12     // Zaehler "Bewegungsstopps"
#define LCP_STATE   15     // letztes Empfangsbyte
#define LCP_DIFF_L  18     // Differenz Motor L
#define LCP_INTEG_L 22     // Integralteil Motor L
#define LCP_ERR     28     // Beginn Fehlerzaehler
#define ERR_SEND    28     // Zaehler Sendversagen 
#define ERR_ETO     29     // Zaehler Empf.-Timeout
#define ERR_WMESS   30     // FZ Winkelmessung 
// Zeichen 31 ist der Zaehler falsche Interrupts


void regelung_reset(){     // Regelung initialisieren
  diffR = 0; 	           // Regelerabweichungen löschen
  diffL = 0;
  integR = 0;              // Integralanteile löschen
  integL = 0;
}

int main(){
  uint8_t state = 'S';     // Anfangszustand gestoppt
  com_pc_init();           // Treiber initialisieren
  rotmess_init();
  pwm_init();
  lcd_init((uint8_t*)INITSTR);
  DDRA = 0;                // Tastereingabe initialisieren
  sei();                   // Interrupts ein

	
  while(1){
    lcd_disp_chr(state, LCP_STATE); // Zustand anzeigen
    if (state=='W'||state=='S'){
      if (com_pc_get(mrmsg)){ // wenn neue Empfangsnachricht
        // neue Sollgeschwindigkeiten, Schrittzahl und Regler-
        speedR = ((int16_t)mrmsg[0] <<8) + mrmsg[1]; // parameter
        speedL = ((int16_t)mrmsg[2] <<8) + mrmsg[3]; // übernehmen
        step_ct= ((uint16_t)mrmsg[4]<<8) + mrmsg[5];
        kp =     ((uint16_t)mrmsg[6]<<8) + mrmsg[7];
        ki =     ((uint16_t)mrmsg[8]<<8) + mrmsg[9];
        if (state=='S'){
          regelung_reset();
          pwm_start();
        }
        state = 'D';          // neuer Zustand "Bewegung"
      }                       // sonst, wenn noch keine
      else if (state == 'W'){ // neue Eingabe Zustand "Weiter"
        pwm_stop();           // Motoren anhalten
        state = 'S';          // "Weiter" => "Stop"
        lcd_disp_val(++stop_ct, LCP_STPCT, 2);
      }
    }
    if ((step_ct>0) && state =='D'){// Zustand "Drive" und Schritte übrig
      int16_t sa, sb;          // temporäre Werte
      if (rotmess_get(&sa, &sb)){ // wenn neuer Messwert
        if (rotmess_err())     // bei Fehler in "rotmess"
          lcd_incErr(ERR_WMESS);// Fehlerzähler erhöhen
        // Ausführung eines PID-Regelschritts je Motor
        diffR  = diffR + speedR - (sa << 8);
        limit(&diffR, 0x100000);
        integR = integR + ((ki * diffR)>>8);
        limit(&integR, 0x400000);
        pwmR = (integR>>8) + ((kp * diffR)>>12);
        limit(&pwmR, 0x4000);
        pwm_set_R(pwmR);
        diffL = diffL + speedL - (sb << 8);
        limit(&diffL, 0x100000);
        integL = integL + ((ki * diffL)>>8);
        limit(&integL, 0x400000);
        pwmL = (integL>>8) + ((kp * diffL)>>12);
        limit(&pwmL, 0x4000);
        pwm_set_L(pwmL);
        step_ct--;
        lcd_disp_val(abs(diffR)>>8, LCP_DIFF_R, 3);
        lcd_disp_val(abs(integR)>>8, LCP_INTEG_R, 3);
        lcd_disp_val(abs(diffL)>>8, LCP_DIFF_L, 3);
        lcd_disp_val(abs(integL)>>8, LCP_INTEG_L, 3);
      }
    }
    if ((step_ct<=0) && state =='D' ){ 
      // bei Bewegungsende oder "Pause" Daten zum PC senden
      msmsg[0] =  diffR>>16; msmsg[1] =  diffR>>8 & 0xFF;
      msmsg[2] = integR>>16; msmsg[3] = integR>>8 & 0xFF;
      msmsg[4] =  diffL>>16; msmsg[5] =  diffL>>8 & 0xFF;
      msmsg[6] = integL>>16; msmsg[7] = integL>>8 & 0xFF;
      if (!com_pc_send(msmsg))  // wenn Senden versagt
        lcd_incErr(ERR_SEND);   // Fehlerzähler erhöhen
      state='W';                // Zustand => "Weiter"
    }
    if (PINA){                  // bei Tastendruck an Port A
      pwm_stop();               // Motoren anhalten
      state = 'S';              // Anfangszustand herstellen
      lcd_disp_str((uint8_t*)"xx", LCP_STPCT, 2);
      lcd_disp_str((uint8_t*)"....", LCP_ERR, 4);
      stop_ct = 0;
    }
  }
}
